package com.inteligr8.activiti.mq;

import org.activiti.engine.delegate.event.ActivitiEntityEvent;
import org.activiti.engine.delegate.event.ActivitiEventType;
import org.activiti.engine.impl.persistence.entity.Entity;

public interface ActivitiEntityEventListener<T extends Entity> {
	
	/**
	 * This method checks to see if this listener is for the specified entity.
	 * 
	 * @param entity An Activiti entity.
	 * @return `true` if this listener supports the entity; `false` otherwise.
	 */
	default boolean ofEntityType(Entity entity) {
		return entity == null ? false : this.ofEntityType(entity.getClass());
	}

	/**
	 * This method checks to see if this listener is for the specified entity
	 * class.
	 * 
	 * @param entityType An Activiti entity class.
	 * @return `true` if this listener supports the entity type; `false` otherwise.
	 */
	boolean ofEntityType(Class<? extends Entity> entityType);

	/**
	 * This method allows for initialization on application startup.  Any
	 * resources this listener should subscribe to, should be connected or
	 * opened.
	 */
	default void onApplicationStartup() {
	}
	
	/**
	 * This method allows for uninitialization on application shutdown.  Any
	 * resources this listener subscribes to should be disconnected or closed.
	 * 
	 * The `ActivitiEntityEventListener#onEvent()` method may still be called
	 * before startup or after shutdown.
	 */
	default void onApplicationShutdown() {
	}
	
	/**
	 * This method is fired every time there is a qualifying entity event.
	 * 
	 * @param aaevent An Activiti entity event.
	 */
	default void onEntityEvent(ActivitiEntityEvent aaevent) {
		@SuppressWarnings("unchecked")
		T entity = (T) aaevent.getEntity();
		
		this.onEntityEvent(aaevent.getType(), aaevent.getProcessDefinitionId(), aaevent.getProcessInstanceId(), aaevent.getExecutionId(), entity);
	}
	
	/**
	 * Thi smethod is fired every time there is a qualifying entity event.
	 * 
	 * @param eventType An Activiti event type; limited to `ENTITY_*`.
	 * @param processDefinitionId The unique identifier of the process definition subject to the event.
	 * @param processInstanceId The unique identifier of the process instance subject to the event.
	 * @param executionId The unique identifier of the execution subject to the event.
	 * @param entity The Activiti entity.
	 */
	void onEntityEvent(ActivitiEventType eventType, String processDefinitionId, String processInstanceId, String executionId, T entity);

}
