package com.inteligr8.activiti.mq;

import java.util.List;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;

import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.event.ActivitiEntityEvent;
import org.activiti.engine.delegate.event.ActivitiEvent;
import org.activiti.engine.delegate.event.ActivitiEventListener;
import org.activiti.engine.delegate.event.ActivitiEventType;
import org.activiti.engine.impl.persistence.entity.Entity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationListener;
import org.springframework.context.event.ApplicationContextEvent;
import org.springframework.context.event.ContextClosedEvent;
import org.springframework.context.event.ContextRefreshedEvent;
import org.springframework.context.event.ContextStartedEvent;
import org.springframework.context.event.ContextStoppedEvent;
import org.springframework.stereotype.Component;

@Component
public class ActivitiEntityEventMonitor implements ActivitiEventListener, ApplicationListener<ApplicationContextEvent> {
	
	private final Logger logger = LoggerFactory.getLogger(ActivitiEntityEventMonitor.class);
	
	@Autowired
	private List<ActivitiEntityEventListener<? extends Entity>> listeners;
	
	@Autowired
	private ProcessEngine services;
	
	/**
	 * This method is fired by Spring, the framework behind Activiti.  This
	 * forwards the application event to each listener so they can initialize
	 * or uninitialize.
	 */
	@Override
	public void onApplicationEvent(ApplicationContextEvent event) {
		if (event instanceof ContextRefreshedEvent || event instanceof ContextStoppedEvent || event instanceof ContextClosedEvent) {
			this.logger.debug("Application context refresh/stop/close detected; shutting down listeners: {}", event);
			
			for (ActivitiEntityEventListener<? extends Entity> listener : this.listeners) {
				listener.onApplicationShutdown();
			}
		}
		
		// the listener cannot be active until the context is initialized
		if (event instanceof ContextRefreshedEvent || event instanceof ContextStartedEvent) {
			this.logger.debug("Application context refresh/start detected; starting up listeners: {}", event);
			
			for (ActivitiEntityEventListener<? extends Entity> listener : this.listeners) {
				listener.onApplicationStartup();
			}
		}
	}
	
	/**
	 * This method will start monitoring for changes in deployment, app, and
	 * process definitions.
	 */
	@PostConstruct
	protected void init() {
		this.logger.debug("Bean initialized; starting to listen for any entity event ...");
		
		this.services.getRuntimeService().addEventListener(this,
				ActivitiEventType.ENTITY_CREATED,
				ActivitiEventType.ENTITY_INITIALIZED,
				ActivitiEventType.ENTITY_UPDATED,
				ActivitiEventType.ENTITY_ACTIVATED,
				ActivitiEventType.ENTITY_SUSPENDED,
				ActivitiEventType.ENTITY_DELETED);
	}
	
	/**
	 * This method will stop monitoring for changes in deployment, app, and
	 * process definitions.
	 */
	@PreDestroy
	protected void uninit() {
		this.logger.debug("Bean uninitialized; stopping listener for any entity event ...");
		
		this.services.getRuntimeService().removeEventListener(this);
	}
	
	/**
	 * This method is fired by the Activiti platform.  It is called on every
	 * entity event.
	 * 
	 * @param event An Activiti event.
	 */
	@Override
	public void onEvent(ActivitiEvent event) {
		this.logger.trace("Triggered by event: {}: {}; pdId: {}; piId: {}; execId: {}", event.getType(), event.getClass(), event.getProcessDefinitionId(), event.getProcessInstanceId(), event.getExecutionId());
		
		for (ActivitiEntityEventListener<? extends Entity> listener : this.listeners) {
			ActivitiEntityEvent aaevent = (ActivitiEntityEvent) event;
			if (aaevent.getEntity() instanceof Entity && listener.ofEntityType((Entity) aaevent.getEntity()))
				listener.onEntityEvent(aaevent);
		}
	}
	
	@Override
	public boolean isFailOnException() {
		return true;
	}

}
