package com.inteligr8.activiti.mq;

import java.util.List;
import java.util.Set;

import org.activiti.engine.delegate.event.ActivitiEventType;
import org.activiti.engine.impl.persistence.entity.DeploymentEntity;
import org.activiti.engine.impl.persistence.entity.Entity;
import org.activiti.engine.impl.persistence.entity.ProcessDefinitionEntity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class MqDeploymentEventListener implements ActivitiEntityEventListener<DeploymentEntity> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	protected MqSubscribeLooper looper;
	
	@Autowired
	protected ProcessDefinitionRegistry registry;
	
	@Autowired
	private MqSubscriptionService subscriptionService;
	
	@Override
	public boolean ofEntityType(Class<? extends Entity> entityType) {
		return DeploymentEntity.class.isAssignableFrom(entityType);
	}
	
	@Override
	public boolean ofEntityType(Entity entity) {
		return entity instanceof DeploymentEntity;
	}
	
	@Override
	public void onEntityEvent(ActivitiEventType eventType, String processDefinitionId, String processInstanceId,
			String executionId, DeploymentEntity entity) {
		this.logger.trace("Triggered by deployment event: {}", entity);
		
		switch (eventType) {
			case ENTITY_INITIALIZED:
				// we cannot use the ProcessDefinitionEntity for ENTITY_INITIALIZED as we need the BpmnModel later and it is not yet cached
				// we must use DeploymentEntity and then dig down for the process definitions
				this.onDeploymentAddEvent((DeploymentEntity) entity);
				break;
			default:
		}
	}
	
	protected void onDeploymentAddEvent(DeploymentEntity entity) {
		this.logger.debug("Triggered by deployment addition: {}", entity);
		
		List<ProcessDefinitionEntity> procDefEntities = entity.getDeployedArtifacts(ProcessDefinitionEntity.class);
		if (procDefEntities == null)
			return;
		this.logger.debug("Found {} process definitions in deployment: {}", procDefEntities.size(), entity.getId());

		for (ProcessDefinitionEntity procDefEntity : procDefEntities) {
			this.logger.debug("Inspecting process definition: {}: {}: {}", procDefEntity.getId(), procDefEntity.getKey(), procDefEntity.getName());

			this.unsubscribeOtherMqSubscribeTasks(procDefEntity.getId());
			
			if (this.registry.isMqStart(procDefEntity.getId()))
				this.looper.loop(procDefEntity.getId());
		}
	}
	
	protected void unsubscribeOtherMqSubscribeTasks(String procDefId) {
		try {
			Set<String> executionIds = this.subscriptionService.cancelAllOtherVersions(procDefId);
			if (this.logger.isDebugEnabled()) {
				this.logger.debug("Subscription starter executions ended early: {}: {}", procDefId, executionIds);
			} else {
				this.logger.info("Subscriptions starter ended early: {}: {}", procDefId, executionIds.size());
			}
		} catch (Exception e) {
			this.logger.error("The subscriptions could not be cancelled: " + procDefId, e);
		}
	}

}
