package com.inteligr8.activiti.mq;

import java.util.Set;

import org.activiti.engine.ActivitiObjectNotFoundException;
import org.activiti.engine.delegate.event.ActivitiEventType;
import org.activiti.engine.impl.persistence.entity.Entity;
import org.activiti.engine.impl.persistence.entity.ProcessDefinitionEntity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class MqProcessDefinitionEventListener implements ActivitiEntityEventListener<ProcessDefinitionEntity> {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	protected MqSubscribeLooper looper;
	
	@Autowired
	protected ProcessDefinitionRegistry registry;
	
	@Autowired
	protected MqSubscriptionService subscriptionService;
	
	@Override
	public boolean ofEntityType(Class<? extends Entity> entityType) {
		return ProcessDefinitionEntity.class.isAssignableFrom(entityType);
	}
	
	@Override
	public boolean ofEntityType(Entity entity) {
		return entity instanceof ProcessDefinitionEntity;
	}
	
	@Override
	public void onApplicationStartup() {
		this.looper.deloopAllMqProcessDefinitions();
		this.looper.loopAllMqProcessDefinitions();
	}
	
	@Override
	public void onApplicationShutdown() {
		this.looper.deloopAllMqProcessDefinitions();
	}
	
	@Override
	public void onEntityEvent(ActivitiEventType eventType, String processDefinitionId, String processInstanceId,
			String executionId, ProcessDefinitionEntity entity) {
		this.logger.trace("Triggered by process definition event: {}", entity);
		
		switch (eventType) {
			case ENTITY_CREATED:
			case ENTITY_INITIALIZED:
				// we cannot use the ProcessDefinitionEntity for ENTITY_CREATED or ENTITY_INITIALIZED as we need the BpmnModel and it is not yet cached
				// we must use DeploymentEntity and then dig down for the process definitions
				break;
			case ENTITY_DELETED:
			case ENTITY_SUSPENDED:
				// we need to stop the listener
				this.onProcessDefinitionRemoveEvent((ProcessDefinitionEntity) entity);
				break;
			default:
		}
		
	}
	
	protected void onProcessDefinitionAddEvent(ProcessDefinitionEntity entity) {
		this.logger.debug("Triggered by process definition addition: {}", entity);

		try {
			// cancel subscriptions on all legacy version process definitions
			this.unsubscribeOtherMqSubscribeTasks(entity.getId());
			
			if (this.registry.isMqStart(entity.getId()))
				this.looper.loop(entity.getId());
		} catch (ActivitiObjectNotFoundException aonfe) {
			this.logger.debug("Added process definition could not be found; because of orders of operation or transaction isolation: {}: {}", entity.getId(), aonfe.getMessage());
		}
	}
	
	protected void onProcessDefinitionRemoveEvent(ProcessDefinitionEntity entity) {
		this.logger.debug("Triggered by process definition removal: {}", entity);

		// cancel subscriptions on this process definition
		this.unsubscribeMqSubscribeTasks(entity.getId());

		if (this.registry.isMqStart(entity.getId()))
			this.looper.deloop(entity.getId());
	}
	
	protected void unsubscribeMqSubscribeTasks(String procDefId) {
		try {
			Set<String> executionIds = this.subscriptionService.cancelAll(procDefId);
			if (this.logger.isDebugEnabled()) {
				this.logger.debug("Subscription executions ended early: {}: {}", procDefId, executionIds);
			} else {
				this.logger.info("Subscriptions ended early: {}: {}", procDefId, executionIds.size());
			}
		} catch (ActivitiObjectNotFoundException aonfe) {
			throw aonfe;
		} catch (Exception e) {
			this.logger.error("The subscriptions could not be cancelled: " + procDefId, e);
		}
	}
	
	protected void unsubscribeOtherMqSubscribeTasks(String procDefId) {
		try {
			Set<String> executionIds = this.subscriptionService.cancelAllOtherVersions(procDefId);
			if (this.logger.isDebugEnabled()) {
				this.logger.debug("Subscription executions ended early: {}: {}", procDefId, executionIds);
			} else {
				this.logger.info("Subscriptions ended early: {}: {}", procDefId, executionIds.size());
			}
		} catch (ActivitiObjectNotFoundException aonfe) {
			throw aonfe;
		} catch (Exception e) {
			this.logger.error("The subscriptions could not be cancelled: " + procDefId, e);
		}
	}

}
