package com.inteligr8.activiti.mq;

import java.io.IOException;
import java.util.concurrent.TimeoutException;

import javax.jms.JMSException;

import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.BpmnError;
import org.activiti.engine.delegate.DelegateExecution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import com.inteligr8.activiti.mq.model.DeliveredMessage;
import com.inteligr8.activiti.mq.model.PreparedMessage;

/**
 * This delegate sends a message to an MQ queue.
 * 
 * The process instance will continue after a successful publish of the message
 * by this delegate to the specified queue.  Use `mqSubscribeReplyDelegate` in
 * a subsequent task to handle the response asynchronously.
 * 
 * If you have more than one MQ communication in a single process, set the
 * `mq_messageName` field on this and the corresponding
 * `mqSubscribeReplyDelegate` task.  The resultant variables will then include
 * that message name (e.g. `mq_messageId_{mq_messageName}`).
 * 
 * @author brian@inteligr8.com
 * @since 1.0
 * @see mqSubscribeReplyDelegate#
 * @see mqSubscribeDelegate#
 */
@Component(MqPublishDelegate.BEAN_ID)
public class MqPublishDelegate extends AbstractMqDelegate {

	public static final String BEAN_ID = "mqPublishDelegate";
    private final Logger logger = LoggerFactory.getLogger(this.getClass());
    
    @Autowired
    private ProcessEngine services;
    
    @Autowired
    private MqServiceTaskService msts;
    
    /**
	 * This method makes this bean an Activiti delegate.
	 * 
	 * @param execution An Activiti delegate execution (source task or execution/task listener).
	 * @field mq_connectorId An Activiti App Tenant Endpoint ID or Java system property in the format `inteligr8.mq.connectors.{connectorId}.url`.  Using system properties support `url`, `username`, and `password`.
	 * @field mq_queueName The name of an MQ destination (queue or topic).  This is the target of the message.  If it doesn't exist, it will be created.
	 * @field mq_metadataProcessScope [optional] `true` to set the `mq_messageId` result variable in the process instance scope; otherwise the variable will be local to the task.  `mq_correlationId` is always at the process scope.
	 * @field mq_messageName [optional] A unique identifier to append to Activiti result variable names when more than one MQ message publication/subscription is supported by a process.
	 * @field mq_priority [optional] A priority of the MQ message.  May be an expression.  Supported range depends on MQ protocol.
	 * @field mq_payload [optional] The body of the MQ message.  May include expressions.
	 * @field mq_payloadMimeType [optional] The MIME type of the `mq_payload`.
	 * @field mq_replyQueueName [optional] The name of an MQ destination (queue or topic).  This tells the processor of the message where to send a reply.
	 * @field mq_statusQueueName [optional] The name of an MQ destination (queue or topic).  This tells the processor of the message where to send status updates.
	 * @varIn mq_correlationId [optional] The correlationId of the message to send.
	 * @varOut mq_correlationId The correlationId of the message sent.
	 * @varOut mq_messageId The messageId of the message sent.
	 * @throws BPMNError timeout The MQ connection timed out connecting or waiting for a message.
	 * @throws BPMNError network The MQ connection experienced network issues.
	 * @throws BPMNError mq An unknown MQ issue occurred.
	 * @see mqSubscribeReplyDelegate#delegate
	 * @see mqSubscribeDelegate#delegate
     */
    @Override
    public void execute(DelegateExecution execution) {
    	MqPublishDelegateExecution mqExecution = new MqPublishDelegateExecution(this.services, this.msts, execution);
    	mqExecution.validate();
    	
    	GenericDestination destination = new GenericDestination();
    	destination.setQueueName(mqExecution.getQueueNameFromModel());
    	
    	try {
    		MqCommunicator communicator = this.getCommunicator(mqExecution.getConnectorIdFromModel());

        	PreparedMessage<String> message = communicator.createPreparedMessage();
        	if (mqExecution.getCorrelationId() != null)
        		message.setCorrelationId(mqExecution.getCorrelationId());
        	if (mqExecution.getStatusQueueNameFromModel() != null)
        		message.setProperty("inteligr8.statusQueueName", mqExecution.getStatusQueueNameFromModel());
        	message.setReplyToQueueName(mqExecution.getReplyQueueNameFromModel());
        	message.setPriority(mqExecution.getPriorityFromModel());
        	
        	String payload = mqExecution.getPayloadFromModel();
        	if (payload != null) {
        		String payloadMimeType = mqExecution.getPayloadMimeTypeFromModel();
        		if (payloadMimeType != null) {
        			message.setContent(payload, payloadMimeType);
        		} else {
        			message.setContent(payload);
        		}
        	}
        	
    		DeliveredMessage<?> deliveredMessage = communicator.send(destination, message);
    		this.logger.debug("Sent MQ message: {} => {} => {} => {}", mqExecution.getConnectorIdFromModel(), mqExecution.getQueueNameFromModel(), deliveredMessage.getMessageId(), deliveredMessage.getCorrelationId());
    		
    		mqExecution.setMessageId(deliveredMessage.getMessageId());
    		mqExecution.setCorrelationId(deliveredMessage.getCorrelationId());
    	} catch (TimeoutException te) {
    		this.logger.error("MQ connection or communication timed out: " + te.getMessage(), te);
    		throw new BpmnError("timeout", "MQ connection or communication timed out: " + te.getMessage());
    	} catch (IOException ie) {
    		this.logger.error("MQ connection or network communication failed: " + ie.getMessage(), ie);
    		throw new BpmnError("network", "MQ connection or network communication failed: " + ie.getMessage());
    	} catch (JMSException je) {
    		this.logger.error("JMS communication failed: " + je.getMessage(), je);
    		throw new BpmnError("mq", "JMS communication failed: " + je.getMessage());
    	}
    }

}
