package com.inteligr8.activiti.mq;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.OverridingMethodsMustInvokeSuper;

import org.activiti.bpmn.model.FieldExtension;
import org.activiti.bpmn.model.ServiceTask;
import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.Expression;
import org.activiti.engine.delegate.VariableScope;
import org.activiti.engine.impl.context.Context;
import org.activiti.engine.impl.el.ExpressionManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class MqServiceTask {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	protected final ProcessEngine services;
	protected final ServiceTask task;
	protected final Map<String, FieldExtension> fieldMap = new HashMap<>();
	private boolean metadataToProcessScope = false;
	
	public MqServiceTask(ProcessEngine services, ServiceTask task) {
		this.services = services;
		this.task = task;

		this.logger.trace("Indexing {} fields", this.task.getFieldExtensions().size());
		for (FieldExtension field : this.task.getFieldExtensions()) {
			this.logger.trace("Discovering field: {}: {}: {}", field.getId(), field.getFieldName(), field.getStringValue());
			
			switch (field.getFieldName()) {
				case Constants.FIELD_METADATA_PROCESS_SCOPE:
					this.metadataToProcessScope = Boolean.valueOf(field.getStringValue());
					break;
				default:
					this.fieldMap.put(field.getFieldName(), field);
			}
		}
	}
	
	@OverridingMethodsMustInvokeSuper
	public void validate() {
    	if (this.fieldMap.get(Constants.FIELD_CONNECTOR_ID) == null)
    		throw new IllegalStateException("The '" + this.task.getId() + "' activity must define an 'MQ Connector ID'");
    	if (this.fieldMap.get(Constants.FIELD_QUEUE_NAME) == null)
    		throw new IllegalStateException("The '" + this.task.getId() + "' activity must define an 'MQ Queue Name'");
	}
	
	public String getTaskId() {
		return this.task.getId();
	}
	
	public boolean doWriteToProcessScope() {
		return this.metadataToProcessScope;
	}
	
	protected <T> T getFieldValueFromModel(String fieldName, Class<T> type) {
		return this.getFieldValueFromModel(fieldName, type, null, false);
	}
	
	protected <T> T getFieldValueFromModel(String fieldName, Class<T> type, VariableScope varscope) {
		return this.getFieldValueFromModel(fieldName, type, varscope, false);
	}
	
	@SuppressWarnings("unchecked")
	protected <T> T getFieldValueFromModel(String fieldName, Class<T> type, VariableScope varscope, boolean forceExpressionProcessing) {
		Object value = this.getFieldValueFromModel(fieldName, varscope, forceExpressionProcessing);
		
		if (value == null) {
			return null;
		} else if (String.class.isAssignableFrom(type)) {
			if (value instanceof String) {
				return (T) value;
			} else {
				return (T) value.toString();
			}
		} else {
			try {
				Method method = type.getMethod("valueOf", value.getClass());
				return (T) method.invoke(null, value);
			} catch (NoSuchMethodException | InvocationTargetException | IllegalAccessException e) {
				String strvalue;
				if (value instanceof String) {
					strvalue = (String) value;
				} else {
					strvalue = value.toString();
				}
				
				try {
					Method method = type.getMethod("valueOf", String.class);
					return (T) method.invoke(null, strvalue);
				} catch (NoSuchMethodException | InvocationTargetException | IllegalAccessException e2) {
					throw new IllegalArgumentException("The target type '" + type + "' has no 'valueOf' method for String or type: " + value.getClass());
				}
			}
		}
	}
	
	protected Object getFieldValueFromModel(String fieldName, VariableScope varscope, boolean forceExpressionProcessing) {
		FieldExtension field = this.fieldMap.get(fieldName);
		if (field == null) {
			return null;
		} else if (field.getExpression() != null && field.getExpression().length() > 0) {
			this.logger.trace("Field value is recognized as an expression by the Activity framework: {}: {}", fieldName, field.getExpression());
			ExpressionManager exprman = Context.getProcessEngineConfiguration().getExpressionManager();
			Expression expr = exprman.createExpression(field.getExpression());
			return expr.getValue(varscope);
		} else if (field.getStringValue() == null) {
			this.logger.trace("Field value is null: {}", fieldName);
			return null;
		} else if (forceExpressionProcessing) {
			this.logger.trace("Field value will be processed as potentially having expression(s): {}", fieldName);
			ExpressionManager exprman = Context.getProcessEngineConfiguration().getExpressionManager();
			Expression expr = exprman.createExpression(field.getStringValue());
			return expr.getValue(varscope);
		} else if (field.getStringValue().startsWith("${") && field.getStringValue().endsWith("}")) {
			this.logger.trace("Field value is recognized as an expression by the MQ extension: {}: {}", fieldName, field.getStringValue());
			ExpressionManager exprman = Context.getProcessEngineConfiguration().getExpressionManager();
			Expression expr = exprman.createExpression(field.getStringValue());
			return expr.getValue(varscope);
		} else {
			return field.getStringValue();
		}
	}

}
