package com.inteligr8.activiti.mq;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.activiti.bpmn.model.FieldExtension;
import org.activiti.bpmn.model.FlowElement;
import org.activiti.bpmn.model.ServiceTask;
import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.event.ActivitiActivityEvent;
import org.activiti.engine.delegate.event.ActivitiEventType;
import org.activiti.engine.repository.ProcessDefinition;
import org.activiti.engine.repository.ProcessDefinitionQuery;
import org.activiti.engine.runtime.Execution;
import org.activiti.engine.runtime.ExecutionQuery;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import com.activiti.domain.idm.Tenant;

@Component
public class MqSubscribeLooper {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	protected ProcessEngine services;
	
	@Autowired
	protected ProcessDefinitionRegistry registry;
	
	@Autowired
	protected TenantFinderService tenantFinderService;
	
	@Autowired
	protected MqSubscriptionService subscriptionService;
	
	protected Map<String, AbstractActivityListener> activeListeners = new HashMap<>();
	
	/**
	 * This method starts the listening of activity completions on all the MQ
	 * subscribe tasks.  This will start the infinite loop where new process
	 * instances are started every time another process instance receives a
	 * message.
	 */
	public void loopAllMqProcessDefinitions() {
		String tenantId = this.findTenantId();
		List<ProcessDefinition> procDefs = this.findLatestActiveProcessDefinnitions(tenantId);
		this.logger.debug("Found {} active process definitions", procDefs.size());
		
		for (ProcessDefinition procDef : procDefs) {
			this.logger.trace("Inspecting process definition for qualifying MQ subscriptions: {}", procDef.getId());
			
			if (this.registry.isMqStart(procDef.getId()))
				this.loop(tenantId, procDef.getId());
		}
	}
	
	/**
	 * This method starts the listening of activity completions on the MQ
	 * subscribe task in the specified process definition.  This will start the
	 * infinite loop where new process instances are started every time another
	 * process instance receives a message.
	 * 
	 * @param processDefinitionId The unique identifier of the process definition to loop.
	 */
	public void loop(String processDefinitionId) {
		String tenantId = this.findTenantId();
		this.loop(tenantId, processDefinitionId);
	}
	
	private void loop(String tenantId, String processDefinitionId) {
		ProcessDefinition procDef = this.services.getRepositoryService().getProcessDefinition(processDefinitionId);
		if (procDef == null)
			throw new IllegalArgumentException("The process definition does not exist: " + processDefinitionId);
		if (procDef.isSuspended())
			throw new IllegalStateException("The process definition is suspended: " + processDefinitionId);
		
		ServiceTask task = this.registry.findMqStartSubscribeTask(processDefinitionId);
		if (task == null)
			throw new IllegalArgumentException("The process definition does not qualify for MQ subscription looping: " + processDefinitionId);
		
		int concurrency = this.determineConcurrency(task);
		this.logger.debug("Process definition MQ subscription task is configured for concurrency: {}: {}", processDefinitionId, concurrency);

		List<Execution> execs = this.findExecutionsByServiceTask(tenantId, processDefinitionId, task);
		this.logger.debug("Process appears to have {} executions waiting on the MQ subscription: {}", execs.size(), processDefinitionId);

		if (execs.size() < concurrency) {
			this.logger.info("Process has {} too few executions waiting on the MQ subscription; starting them: {}", (concurrency - execs.size()), processDefinitionId);
			this.loop(tenantId, processDefinitionId, task, concurrency, execs);
		}
	}
	
	private synchronized void loop(String tenantId, String processDefinitionId, ServiceTask task, int concurrency, List<Execution> execs) {
		String key = processDefinitionId + "|" + task.getId();
		if (this.activeListeners.containsKey(key))
			// one listener, no matter how many instances are subscribed
			return;
		
		AbstractActivityListener listener = new AbstractActivityListener(processDefinitionId, task) {
			@Override
			public void onEvent(ActivitiActivityEvent event) {
				logger.debug("MQ message received; starting another process instance on process '{}' to re-subscribe to the MQ queue", event.getProcessDefinitionId());
				services.getRuntimeService().startProcessInstanceById(event.getProcessDefinitionId());
			}
			
			@Override
			public boolean isFailOnException() {
				return true;
			}
		};
		
		// the ServiceTask will then execute, waiting for a message on the queue
		// when a message is received, it will be responsible for starting a new process instance
		// that new process instance will wait for the next message on the queue
		// repeat
		this.services.getRuntimeService().addEventListener(listener, ActivitiEventType.ACTIVITY_COMPLETED);
		
		// register the listener so we can possibly remove it later
		this.activeListeners.put(key, listener);
		
		for (int thread = execs.size(); thread < concurrency; thread++) {
			// start a process instance on the process
			this.logger.debug("Starting process instance on process '{}' to subscribe to an MQ queue/topic", processDefinitionId);
			this.services.getRuntimeService().startProcessInstanceById(processDefinitionId);
		}
	}
	
	/**
	 * This method stops the listening of activity completions.  This will end
	 * the infinite loop where new process instances are started every time
	 * another process instance receives a message.
	 * 
	 * It will also attempt to cancel/close any active MQ subscription.  If
	 * they are already closed, it will be skipped.  The `executionId` of
	 * successfully cancelled subscriptions will be returned. 
	 * 
	 * @return A set of unique identifiers for Activiti executions that were successfully cancelled.
	 */
	public synchronized Set<String> deloopAllMqProcessDefinitions() {
		Set<String> executionIds = new HashSet<>();
		
		Iterator<Entry<String, AbstractActivityListener>> i = this.activeListeners.entrySet().iterator();
		while (i.hasNext()) {
			Entry<String, AbstractActivityListener> listener = i.next();

			String[] splitkey = listener.getKey().split("\\|");
			String processDefinitionId = splitkey[0];
			String activityId = splitkey[1];
			
			executionIds.addAll(this.deloop(processDefinitionId, activityId, listener.getValue()));
			
			i.remove();
		}
		
		return executionIds;
	}
	
	/**
	 * This method stops the listening of activity completions on the MQ
	 * subscribe task in the specified process definition.  This will end the
	 * infinite loop where new process instances are started every time another
	 * process instance receives a message.
	 * 
	 * @param processDefinitionId The unique identifier of the process definition to loop.
	 * @return A set of unique identifiers for Activiti executions that were successfully cancelled.
	 */
	public synchronized Set<String> deloop(String processDefinitionId) {
		ServiceTask task = this.registry.findMqStartSubscribeTask(processDefinitionId);
		if (task == null)
			throw new IllegalArgumentException();
		
		String key = processDefinitionId + "|" + task.getId();
		AbstractActivityListener listener = this.activeListeners.remove(key);
		return this.deloop(processDefinitionId, task.getId(), listener);
	}
	
	private Set<String> deloop(String processDefinitionId, String activityId, AbstractActivityListener listener) {
		if (listener == null) {
			this.logger.trace("No MQ subscription listener was found; no listener to stop");
		} else {
			this.logger.debug("An MQ subscription listener was found; removing/stopping: {}", processDefinitionId);
			this.services.getRuntimeService().removeEventListener(listener);
		}

		Set<String> executionIds = new HashSet<>();
		
		List<Execution> executions = this.services.getRuntimeService().createExecutionQuery()
				.processDefinitionId(processDefinitionId)
				.activityId(activityId)
				.list();
		for (Execution execution : executions) {
			this.logger.debug("An MQ subscription was found; stopping execution: {}: {}: {}", processDefinitionId, activityId, execution.getId());
			if (this.subscriptionService.cancelled(execution)) {
				this.logger.info("An MQ subscription was closed: {}", execution.getId());
				executionIds.add(execution.getId());
			} else {
				this.logger.debug("An MQ subscription was already closed: {}", execution.getId());
			}
		}
		
		return executionIds;
	}
	
	protected String findTenantId() {
		Tenant tenant = this.tenantFinderService.findTenant();
		return this.tenantFinderService.transform(tenant);
	}
	
	protected List<ProcessDefinition> findLatestActiveProcessDefinnitions(String tenantId) {
		ProcessDefinitionQuery procDefQuery = this.services.getRepositoryService().createProcessDefinitionQuery()
				.latestVersion()
				.active();
		if (tenantId == null) {
			procDefQuery.processDefinitionWithoutTenantId();
		} else {
			procDefQuery.processDefinitionTenantId(tenantId);
		}
		
		return procDefQuery.list();
	}
	
	protected List<Execution> findExecutionsByServiceTask(String tenantId, String processDefinitionId, ServiceTask task) {
		ExecutionQuery execQuery = this.services.getRuntimeService().createExecutionQuery()
				.processDefinitionId(processDefinitionId)
				.activityId(task.getId());
		if (tenantId == null) {
			execQuery.executionWithoutTenantId();
		} else {
			execQuery.executionTenantId(tenantId);
		}
		
		return execQuery.list();
	}
	
	protected Integer findConcurrency(ServiceTask task) {
		for (FieldExtension fieldext : task.getFieldExtensions()) {
			if (fieldext.getFieldName().equals(Constants.FIELD_CONCURRENCY)) {
				String concurrencyStr = StringUtils.trimToNull(fieldext.getStringValue());
				return concurrencyStr == null ? null : Integer.valueOf(concurrencyStr);
			}
		}
		
		return null;
	}
	
	protected int determineConcurrency(ServiceTask task) {
		Integer concurrency = this.findConcurrency(task);
		if (concurrency == null) {
			return 1;
		} else if (concurrency.intValue() < 1) {
			this.logger.warn("The task defines an illegal concurrency of {}; using 1: {}", concurrency, task.getId());
			return 1;
		} else {
			return concurrency.intValue();
		}
	}
	
	protected ServiceTask castToServiceTask(FlowElement flowElement) {
		if (!(flowElement instanceof ServiceTask))
			return null;
		return (ServiceTask) flowElement;
	}

}
