package com.inteligr8.activiti.mq;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.DelegateExecution;
import org.activiti.engine.repository.ProcessDefinition;
import org.activiti.engine.runtime.Execution;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class MqSubscriptionService extends MqExecutionService {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private ProcessEngine services;
    
    /**
     * The size of the map has no limit.  It will grow with the number of
     * executions (related to process instances).
     * 
     * This means the map needs to be trimmed.  When an MQ subscribe task is
     * completed, it is paramount to remove the execution from the map.  It is
     * also necessary to remove the execution when it is removed from the
     * `activityExecutionMap` map.
     */
    private Map<String, Pair<Boolean, AutoCloseable>> executionSubscriptionMap = new HashMap<>();

    /**
     * This method registers a consuming execution; an execution that was not
     * automatically started as part of the framework.  This is for reply
     * subscriptions and not starter subscriptions.
     * 
     * @param execution An Activiti execution.
     * @param consumerCloseable A closeable MQ consumer.
     */
    public synchronized void consuming(DelegateExecution execution, AutoCloseable consumerCloseable) {
    	this.executing(execution);
    	this.executionSubscriptionMap.put(execution.getId(), Pair.of(false, consumerCloseable));
    }
    
    /**
     * This method registers an auto-consuming execution; an execution that was
     * automatically started as part of the framework.  This is for starter
     * subscriptions and not reply subscriptions.
     * 
     * @param execution An Activiti execution.
     * @param consumerCloseable A closeable MQ consumer.
     */
    public synchronized void autoconsuming(DelegateExecution execution, AutoCloseable consumerCloseable) {
    	this.executing(execution);
    	this.executionSubscriptionMap.put(execution.getId(), Pair.of(true, consumerCloseable));
    }
    
    public synchronized void consumed(DelegateExecution execution, AutoCloseable consumerCloseable) {
    	Pair<Boolean, AutoCloseable> cachedConsumerCloseable = this.executionSubscriptionMap.get(execution.getId());
    	if (cachedConsumerCloseable != null && cachedConsumerCloseable.getRight() != consumerCloseable)
    		throw new IllegalStateException("The consumer objects were expected to be identical");
    	
    	this.executionSubscriptionMap.remove(execution.getId());
    	this.executed(execution);
    }
    
    @Override
    public synchronized boolean cancelled(Execution execution) {
    	Pair<Boolean, AutoCloseable> cachedConsumerCloseable = this.executionSubscriptionMap.get(execution.getId());
    	if (cachedConsumerCloseable == null) {
    		this.logger.trace("An execution was cancelled, but had no registered subscription to close: {}", execution.getId());
    		return false;
    	}
    	
    	// this will eventually lead to a call to "consumed() above"
    	try {
    		cachedConsumerCloseable.getRight().close();
    		return super.cancelled(execution);
    	} catch (Exception e) {
    		throw new IllegalStateException(e);
    	}
    }
    
    /**
     * This method cancel the MQ subscription on the specified execution.
     * 
     * @param executionId An execution unique identifier.
     * @return `true` if execution and MQ subscription were ended; `false` otherwise.
     */
    @Override
    public synchronized boolean cancel(String executionId) throws Exception {
    	if (!super.cancel(executionId))
    		return false;
    	
		this.logger.trace("Removing MQ subscription execution: {}", executionId);
		Pair<Boolean, AutoCloseable> consumer = this.executionSubscriptionMap.remove(executionId);
		if (consumer != null) {
    		this.logger.debug("Closing MessageConsumer to terminate a subscription early: {}: {}", executionId, consumer.getRight());
			consumer.getRight().close();
		}
		
		return true;
    }
    
    /**
     * This method cancels all the MQ subscriptions active on the specified
     * process definition.
     * 
     * @param processDefinitionId A process definition version unique identifier.
     * @return A set of execution identifiers subscribed to MQ that are now cancelled; all subscriptions now ended. 
     */
    @Override
    public synchronized Set<String> cancelAll(String processDefinitionId) throws Exception {
    	Set<String> executionIds = super.cancelAll(processDefinitionId);

    	ProcessDefinition processDefinition = this.services.getRepositoryService().getProcessDefinition(processDefinitionId);
    	String processDefinitionKey = processDefinition.getKey();
    	
    	for (String executionId : executionIds) {
    		Pair<Boolean, AutoCloseable> consumer = this.executionSubscriptionMap.remove(executionId);
    		if (consumer != null && consumer.getLeft()) {
        		this.logger.trace("Removing MQ subscription execution: {}: {}", processDefinitionId, executionId);
        		this.logger.debug("Closing MessageConsumer to terminate a subscription early: {}: {}: {}", processDefinitionKey, executionId, consumer.getRight());
    			consumer.getRight().close();
    		}
    	}
    	
    	return executionIds;
    }
    
    /**
     * @param latestProcessDefinitionId A process definition identifier to NOT clear.  All other versions will be cleared.
     * @return A set of execution identifiers subscribed to MQ that were in the now cleared map; all subscriptions now ended. 
     */
    @Override
    public synchronized Set<String> cancelAllOtherVersions(String latestProcessDefinitionId) throws Exception {
    	Set<String> executionIds = super.cancelAllOtherVersions(latestProcessDefinitionId);

    	ProcessDefinition latestProcessDefinition = this.services.getRepositoryService().getProcessDefinition(latestProcessDefinitionId);
    	String processDefinitionKey = latestProcessDefinition.getKey();
    	
    	for (String executionId : executionIds) {
    		Pair<Boolean, AutoCloseable> consumer = this.executionSubscriptionMap.remove(executionId);
    		if (consumer != null && consumer.getLeft()) {
        		this.logger.trace("Removing MQ subscription execution: {}: {}", latestProcessDefinitionId, executionId);
        		this.logger.debug("Closing MessageConsumer to terminate a subscription early: {}: {}: {}", processDefinitionKey, executionId, consumer.getRight());
    			consumer.getRight().close();
    		}
    	}
    	
    	return executionIds;
    }

}
