package com.inteligr8.activiti.mq;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.annotation.PostConstruct;

import org.activiti.bpmn.model.BpmnModel;
import org.activiti.bpmn.model.FlowElement;
import org.activiti.bpmn.model.SequenceFlow;
import org.activiti.bpmn.model.ServiceTask;
import org.activiti.bpmn.model.StartEvent;
import org.activiti.engine.ActivitiObjectNotFoundException;
import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.JavaDelegate;
import org.activiti.engine.impl.bpmn.behavior.NoneStartEventActivityBehavior;
import org.activiti.engine.impl.context.Context;
import org.activiti.engine.impl.util.ProcessDefinitionUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

@Component
public class ProcessDefinitionRegistry {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final Pattern expressionPattern = Pattern.compile("\\$\\{(.+)\\}");
	
	@Autowired
	protected ProcessEngine services;
	
	@Autowired
	protected ApplicationContext context;
	
	private Map<String, ServiceTask> processDefinitionMqSubscribeTasks;
	
	@PostConstruct
	private void init() {
		this.processDefinitionMqSubscribeTasks = new HashMap<>();
	}
	
	public synchronized boolean isMqStart(String processDefinitionId) {
		if (this.processDefinitionMqSubscribeTasks.containsKey(processDefinitionId)) {
			return this.processDefinitionMqSubscribeTasks.get(processDefinitionId) != null;
		} else {
			try {
				// not yet cached; cache it; then return
				return this.findMqStartSubscribeTask(processDefinitionId) != null;
			} catch (ActivitiObjectNotFoundException aonfe) {
				return false;
			}
		}
	}
	
	public synchronized ServiceTask findMqStartSubscribeTask(String processDefinitionId) {
		ServiceTask task = this.processDefinitionMqSubscribeTasks.get(processDefinitionId);
		if (task != null)
			return task;
		
		BpmnModel model = this.getBpmnModel(processDefinitionId);
		StartEvent startElement = this.findMqPlainStartEvent(processDefinitionId, model);
		if (startElement == null)
			return null;
		
		List<SequenceFlow> flows = startElement.getOutgoingFlows();
		this.logger.trace("Found {} outgoing flows in the process start element: {}: {}", flows.size(), processDefinitionId, startElement.getId());
		if (flows.isEmpty()) {
			this.logger.warn("A start event is expected to always have an outgoing flow; skipping process: {}", processDefinitionId);
			return null;
		} else if (flows.size() > 1) {
			this.logger.debug("A start event with multiple outgoing flows cannot have an MQ subscription; skipping process: {}", processDefinitionId);
			return null;
		}
		
		SequenceFlow flow = flows.iterator().next();
		FlowElement targetFlowElement = flow.getTargetFlowElement();
		if (targetFlowElement == null) {
			this.logger.warn("A start event outgoing flow is expected to always have a target element; skipping process: {}", processDefinitionId);
			return null;
		}
		this.logger.trace("Found element '{}' after start event '{}' in process '{}'", targetFlowElement.getId(), startElement.getId(), processDefinitionId);

		if (!(targetFlowElement instanceof ServiceTask)) {
			this.logger.trace("A non-service task immediately after start event; skipping process: {}: {}: {}", processDefinitionId, targetFlowElement.getId(), targetFlowElement.getName());
			return null;
		}
		
		task = (ServiceTask) targetFlowElement;
		if (!"delegateExpression".equals(task.getImplementationType())) {
			this.logger.trace("A non-delegate service task immediately after start event; skipping process: {}: {}: {}", processDefinitionId, task.getId(), task.getName());
			return null;
		}
		this.logger.trace("Found service task immediately after start event: {}: {}", task.getImplementationType(), task.getImplementation());
		
		Matcher matcher = this.expressionPattern.matcher(task.getImplementation());
		if (!matcher.find()) {
			this.logger.warn("A service task delegate expression is in an unexpected format: {}; skipping process: {}: {}: {}", task.getImplementation(), processDefinitionId, task.getId(), task.getName());
			return null;
		}
		
		String beanId = matcher.group(1).trim();
		this.logger.trace("Looking up bean: {}", beanId);
		
		JavaDelegate delegateBean = this.context.getBean(beanId, JavaDelegate.class);
		this.logger.trace("Found bean: {}: {}", beanId, delegateBean);
		if (delegateBean == null) {
			this.logger.trace("The service task delegate has no bean; skipping process: {}: {}: {}", processDefinitionId, task.getId(), task.getName());
			return null;
		} else if (!(delegateBean instanceof MqSubscribeDelegate)) {
			this.logger.trace("The service task delegate is not an MQ subscription; skipping process: {}: {}: {}", processDefinitionId, task.getId(), task.getName());
			return null;
		}
		
		this.processDefinitionMqSubscribeTasks.put(processDefinitionId, task);
		this.logger.info("Process starts with an MQ subscription: {}: {}", processDefinitionId, task.getId(), task.getName());
		return task;
	}
	
	private BpmnModel getBpmnModel(String processDefId) {
		if (Context.getProcessDefinitionHelper() == null) {
			// typically during initial startup
			return this.services.getRepositoryService().getBpmnModel(processDefId);
		} else {
			// we cannot use the RepositoryService to get the BpmnModel as it is not yet cached in TX
			return ProcessDefinitionUtil.getBpmnModel(processDefId);
		}
	}
	
	protected StartEvent findMqPlainStartEvent(String processDefId, BpmnModel model) {
		this.logger.trace("Finding all start elements in process: {}", processDefId);
		List<StartEvent> startElements = model.getMainProcess().findFlowElementsOfType(StartEvent.class);
		this.logger.trace("Found {} start elements in process: {}", startElements.size(), processDefId);
		for (StartEvent startElement : startElements) {
			// constrain to just 0..1 plain start events; ignore timer, signal, message, and error start events
			if (!(startElement.getBehavior() instanceof NoneStartEventActivityBehavior)) {
				this.logger.trace("Found non-plain start event; ignoring start event: {}: {}: {}", processDefId, startElement.getId(), startElement.getBehavior());
				continue;
			}
			
			return startElement;
		}
		
		return null;
	}

}
