package com.inteligr8.activiti.mq.amqp;

import java.io.IOException;
import java.net.URISyntaxException;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.Collections;
import java.util.UUID;
import java.util.concurrent.TimeoutException;

import javax.jms.JMSException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.inteligr8.activiti.mq.GenericDestination;
import com.inteligr8.activiti.mq.MqCommunicator;
import com.inteligr8.activiti.mq.MqSubscriptionListener;
import com.inteligr8.activiti.mq.TransactionalMessageHandler;
import com.inteligr8.activiti.mq.model.DeliveredMessage;
import com.inteligr8.activiti.mq.model.PreparedMessage;
import com.rabbitmq.client.AMQP;
import com.rabbitmq.client.CancelCallback;
import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;
import com.rabbitmq.client.DeliverCallback;
import com.rabbitmq.client.Delivery;

public class AmqpCommunicator implements MqCommunicator {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final ConnectionFactory confactory;
	
	public AmqpCommunicator(String url, String username, String password) {
    	this.confactory = new ConnectionFactory();
    	this.confactory.setUsername(username);
    	this.confactory.setPassword(password);
    	try {
    		this.confactory.setUri(url);
    	} catch (KeyManagementException kme) {
    		throw new RuntimeException(kme);
    	} catch (NoSuchAlgorithmException nsae) {
    		throw new RuntimeException(nsae);
    	} catch (URISyntaxException use) {
    		throw new RuntimeException(use);
    	}
	}
	
	@Override
	public boolean validateConnection() {
		try {
			Connection con = this.connect();
			try {
				return true;
			} finally {
				con.close();
			}
		} catch (TimeoutException | IOException e) {
			this.logger.debug("Connection found to be invalid: " + e.getMessage(), e);
			return false;
		}
	}
	
	@Override
	public <BodyType> PreparedMessage<BodyType> createPreparedMessage() {
		return new AmqpPreparedMessage<>();
	}
	
	public Connection connect() throws TimeoutException, IOException {
		return this.confactory.newConnection();
	}
	
	@Override
	public <BodyType> DeliveredMessage<BodyType> send(GenericDestination destination, PreparedMessage<BodyType> message) throws JMSException, IOException, TimeoutException {
		String correlationId = message.getCorrelationId() != null ? message.getCorrelationId() : UUID.randomUUID().toString();
		
		AMQP.BasicProperties bprops = new AMQP.BasicProperties(null, null,
				message.getProperties(),
				destination.getDeliveryMode() != null ? destination.getDeliveryMode() : null,
				destination.getPriority() != null ? destination.getPriority() : null,
				correlationId,
				message.getReplyToQueueName() != null ? message.getReplyToQueueName() : null,
				null, // expiration
				UUID.randomUUID().toString(), // messageId
				null, // timestamp
				null, // type
				null, null, null);
		
		Connection con = this.connect();
		try {
			Channel channel = con.createChannel();
			try {
				channel.queueDeclare(destination.getQueueName(), true, false, false, Collections.emptyMap());
				channel.basicPublish(destination.getQueueName(), null, bprops, null);
			} finally {
				channel.close();
			}
		} finally {
			con.close();
		}
		
		return AmqpDeliveredMessage.transform(bprops);
	}
	
	@Override
	public <BodyType> DeliveredMessage<BodyType> receive(GenericDestination destination, long timeoutInMillis, final String correlationId, MqSubscriptionListener listener, TransactionalMessageHandler<BodyType> handler) throws JMSException, IOException, TimeoutException {
		Connection con = this.connect();
		try {
			Channel channel = con.createChannel();
			try {
				if (listener != null)
					listener.consuming(channel);
				
				channel.queueDeclare(destination.getQueueName(), true, false, false, Collections.emptyMap());
				channel.basicConsume(
						destination.getQueueName(),
						false, // auto-ack
						new DeliverCallback() {
							@Override
							public void handle(String consumerTag, Delivery message) throws IOException {
								// FIXME we need a better solution here; maybe one consumer that distributes them baced on correlationId
								if (correlationId != null && !correlationId.equals(message.getProperties().getCorrelationId()))
									channel.basicNack(message.getEnvelope().getDeliveryTag(), true, true);
								
								logger.debug("Received message '{}' from queue: {}", message.getProperties().getMessageId(), message.getEnvelope().getExchange());
								// TODO
							}
						},
						new CancelCallback() {
							@Override
							public void handle(String consumerTag) throws IOException {
							}
						}
				);
			} finally {
				if (listener != null)
					listener.consumed(channel);
				
				channel.close();
			}
		} finally {
			con.close();
		}
		
		return null;
	}

}
