package com.inteligr8.activiti.mq.jms;

import java.io.IOException;
import java.util.concurrent.TimeoutException;

import javax.jms.Connection;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Session;

import org.apache.activemq.ActiveMQConnectionFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.inteligr8.activiti.mq.GenericDestination;
import com.inteligr8.activiti.mq.MqCommunicator;
import com.inteligr8.activiti.mq.MqSubscriptionListener;
import com.inteligr8.activiti.mq.TransactionalMessageHandler;
import com.inteligr8.activiti.mq.model.DeliveredMessage;
import com.inteligr8.activiti.mq.model.PreparedMessage;

public class JmsCommunicator implements MqCommunicator {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final ActiveMQConnectionFactory confactory;
	
	public JmsCommunicator(String url, String username, String password) throws JMSException {
    	this.confactory = new ActiveMQConnectionFactory();
    	this.confactory.setUserName(username);
    	this.confactory.setPassword(password);
		this.confactory.setBrokerURL(url);
		this.confactory.setClientIDPrefix("activiti-app");
	}
	
	@Override
	public boolean validateConnection() {
		try {
			Connection con = this.connect();
			try {
				Session session = this.start(con);
				try {
					return true;
				} finally {
					session.close();
				}
			} finally {
				con.close();
			}
		} catch (JMSException je) {
			this.logger.debug("Connection found to be invalid: " + je.getMessage(), je);
			return false;
		}
	}
	
	@Override
	public <BodyType> PreparedMessage<BodyType> createPreparedMessage() {
		return new JmsPreparedMessage<>();
	}
	
	public Connection connect() throws JMSException {
		return this.confactory.createConnection();
	}
	
	public Session start(Connection con) throws JMSException {
		return con.createSession(Session.SESSION_TRANSACTED);
	}
	
	@Override
	public <BodyType> DeliveredMessage<BodyType> send(GenericDestination destination, PreparedMessage<BodyType> message) throws JMSException {
		Connection con = this.connect();
		try {
			return this.send(con, destination, (JmsPreparedMessage<BodyType>) message);
		} finally {
			con.close();
		}
	}
	
	protected <BodyType> DeliveredMessage<BodyType> send(Connection con, GenericDestination destination, JmsPreparedMessage<BodyType> message) throws JMSException {
		Session session = this.start(con);
		try {
			DeliveredMessage<BodyType> dmessage = this.send(session, destination, message);
			session.commit();
			return dmessage;
		} finally {
			session.close();
		}
	}
	
	public <BodyType> DeliveredMessage<BodyType> send(Session session, GenericDestination destination, JmsPreparedMessage<BodyType> message) throws JMSException {
		MessageProducer messenger = session.createProducer(destination.toJmsQueue(session));
		try {
			if (destination.getDeliveryMode() != null)
				messenger.setDeliveryMode(destination.getDeliveryMode());
			if (destination.getPriority() != null)
				messenger.setPriority(destination.getPriority());
			if (destination.getTimeToLive() != null)
				messenger.setTimeToLive(destination.getTimeToLive());
			
			Message jmsmsg = message.transform(session);

			this.logger.debug("Sending message to queue: {}", destination.getQueueName());
			messenger.send(jmsmsg);
			
			return JmsDeliveredMessage.transform(jmsmsg);
		} finally {
			messenger.close();
		}
	}
	
	@Override
	public <BodyType> DeliveredMessage<BodyType> receive(GenericDestination destination, long timeoutInMillis, String correlationId, MqSubscriptionListener listener, TransactionalMessageHandler<BodyType> handler) throws JMSException, TimeoutException, IOException {
		Connection con = this.connect();
		try {
			return this.receive(con, destination, timeoutInMillis, correlationId, listener, handler);
		} finally {
			con.close();
		}
	}
	
	protected <BodyType> JmsDeliveredMessage<BodyType> receive(Connection con, GenericDestination destination, long timeoutInMillis, String correlationId, MqSubscriptionListener listener, TransactionalMessageHandler<BodyType> handler) throws JMSException, TimeoutException, IOException {
		con.start();
		try {
			Session session = this.start(con);
			try {
				JmsDeliveredMessage<BodyType> message = this.receive(session, destination, timeoutInMillis, correlationId, listener);
				if (message == null)
					return null;
				
				if (handler != null)
					handler.onMessage(message);
				
				session.commit();
				return message;
			} finally {
				session.close();
			}
		} finally {
			con.stop();
		}
	}
	
	public <BodyType> JmsDeliveredMessage<BodyType> receive(Session session, GenericDestination destination, long timeoutInMillis, String correlationId, MqSubscriptionListener listener) throws JMSException, TimeoutException {
		String messageSelector = correlationId == null ? null : ("JMSCorrelationID='" + correlationId + "'");
		Message receivedMessage = null;
		
		MessageConsumer messenger = session.createConsumer(destination.toJmsQueue(session), messageSelector);
		try {
			if (listener != null)
				listener.consuming(messenger);
			try {
				long startTime = System.currentTimeMillis();
				if (timeoutInMillis < 0L) {
					this.logger.debug("Waiting for message indefinitely: {}", destination.getQueueName());
					receivedMessage = messenger.receive();
				} else if (timeoutInMillis == 0L) {
					this.logger.debug("Checking for message without waiting: {}", destination.getQueueName());
					receivedMessage = messenger.receiveNoWait();
				} else {
					this.logger.debug("Waiting for message for {} ms: {}", timeoutInMillis, destination.getQueueName());
					receivedMessage = messenger.receive(timeoutInMillis);
				}
	
				long elapsedTime = System.currentTimeMillis() - startTime;
				if (receivedMessage != null) {
					this.logger.debug("Received message after {} ms: {}", elapsedTime, destination.getQueueName());
					return JmsDeliveredMessage.transform(receivedMessage);
				}
	
				if (timeoutInMillis > 0L) {
					if (elapsedTime >= timeoutInMillis)
						throw new TimeoutException("A timeout of " + timeoutInMillis + " ms was reached");
					this.logger.debug("Done waiting for message after {} minutes: {}", elapsedTime / 60000L, destination.getQueueName());
				}
				
				return null;
			} finally {
				if (listener != null)
					listener.consumed(messenger);
			}
		} finally {
			messenger.close();
		}
	}
}
