package com.inteligr8.activiti.mq.jms;

import java.io.IOException;
import java.io.InputStream;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.Enumeration;

import javax.jms.BytesMessage;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Queue;
import javax.jms.StreamMessage;
import javax.jms.TextMessage;
import javax.jms.Topic;

import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.inteligr8.activiti.mq.model.AbstractMessage;
import com.inteligr8.activiti.mq.model.DeliveredMessage;

public class JmsDeliveredMessage<BodyType> extends AbstractMessage implements DeliveredMessage<BodyType> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@SuppressWarnings("unchecked")
	public JmsDeliveredMessage(Message message, Class<BodyType> bodyType) throws JMSException {
		this.setCorrelationId(message.getJMSCorrelationID());
		this.setMessageId(message.getJMSMessageID());
		this.setPriority(message.getJMSPriority());
		
		if (message.getJMSDeliveryTime() > 0L) {
			this.setDeliveryTime(OffsetDateTime.ofInstant(Instant.ofEpochMilli(message.getJMSDeliveryTime()), ZoneOffset.UTC));
		} else if (message.getJMSTimestamp() > 0L) {
			this.setDeliveryTime(OffsetDateTime.ofInstant(Instant.ofEpochMilli(message.getJMSTimestamp()), ZoneOffset.UTC));
		}
		
		if (message instanceof TextMessage) {
			if (bodyType == null || String.class.isAssignableFrom(bodyType)) {
				this.setContent((BodyType) message.getBody(String.class));
			} else {
				throw new IllegalArgumentException();
			}
		} else if (message instanceof BytesMessage) {
			if (bodyType == null || bodyType.isArray() && Byte.class.equals(bodyType.getComponentType())) {
				InputStream istream = new BytesMessageInputStream((BytesMessage) message);
				try {
					this.setContent((BodyType) IOUtils.toByteArray(istream));
				} catch (IOException ie) {
					throw (JMSException) ie.getCause();
				}
			} else {
				throw new IllegalArgumentException();
			}
		} else if (message instanceof StreamMessage) {
			if (bodyType == null || InputStream.class.isAssignableFrom(bodyType)) {
				InputStream istream = new StreamMessageInputStream((StreamMessage) message);
				this.setContent((BodyType) istream);
			} else {
				throw new IllegalArgumentException();
			}
		} else {
			// message with no content
			String body = message.getBody(String.class);
			if (body !=  null && body.length() > 0) {
				this.logger.warn("A message with an unexpected content body was received: {}", message.getJMSMessageID());
				this.logger.debug("Message '{}' body: {}", message.getJMSMessageID(), body);
			}
			
			this.setContent((BodyType) body);
		}
		
		Destination destination = message.getJMSReplyTo();
		if (destination instanceof Queue) {
			this.setReplyToQueueName(((Queue) destination).getQueueName());
		} else if (destination instanceof Topic) {
			this.setReplyToQueueName(((Topic) destination).getTopicName());
		}
		
		Enumeration<String> keys = message.getPropertyNames();
		while (keys.hasMoreElements()) {
			String key = keys.nextElement();
			this.setProperty(key, message.getObjectProperty(key));
		}
	}
	
	public static <BodyType> JmsDeliveredMessage<BodyType> transform(Message message) throws JMSException {
		if (message == null)
			return null;
		return new JmsDeliveredMessage<BodyType>(message, null);
	}
	
	public static <BodyType> JmsDeliveredMessage<BodyType> transform(Message message, Class<BodyType> bodyType) throws JMSException {
		if (message == null)
			return null;
		return new JmsDeliveredMessage<BodyType>(message, bodyType);
	}

}
