package com.inteligr8.activiti.mq;

import java.io.IOException;
import java.net.URI;
import java.time.OffsetDateTime;
import java.util.concurrent.TimeoutException;

import javax.annotation.PostConstruct;
import javax.jms.JMSException;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import com.inteligr8.activiti.mq.model.DeliveredMessage;
import com.inteligr8.activiti.mq.model.PreparedMessage;

@Component
public class MQDockerIT {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private MqCommunicationService commsService;
	
	@PostConstruct
	private void init() throws JMSException, IOException, TimeoutException, InterruptedException {
		this.testDockerTcpConnection();
		this.testDockerTcpReceiveEmpty();
		this.testDockerTcpSendReceive();
		this.testDockerFailoverConnection();
		this.testDockerFailoverReceiveEmpty();
		this.testDockerFailoverSendReceive();
	}

	@Test
	public void testDockerTcpConnection() throws JMSException {
		this.testDockerConnection("tcp://mq-activiti-ext-mq:61616");
	}

	@Test
	public void testDockerTcpReceiveEmpty() throws JMSException, IOException {
		this.testDockerReceiveEmpty("tcp://mq-activiti-ext-mq:61616");
	}

	@Test
	public void testDockerTcpSendReceive() throws JMSException, IOException, TimeoutException, InterruptedException {
		this.testDockerSendReceive("tcp://mq-activiti-ext-mq:61616");
	}
	
	@Test
	public void testDockerFailoverConnection() throws JMSException {
		this.testDockerConnection("failover:(tcp://mq-activiti-ext-mq:61616)?timeout=5000");
	}

	@Test
	public void testDockerFailoverReceiveEmpty() throws JMSException, IOException {
		this.testDockerReceiveEmpty("failover:(tcp://mq-activiti-ext-mq:61616)?timeout=5000");
	}

	@Test
	public void testDockerFailoverSendReceive() throws JMSException, IOException, TimeoutException, InterruptedException {
		this.testDockerSendReceive("failover:(tcp://mq-activiti-ext-mq:61616)?timeout=5000");
	}

	protected void testDockerConnection(String url) throws JMSException {
		this.logger.info("Testing MQ connection: {}", url);
		MqCommunicator communicator = this.commsService.getCommunicator(URI.create(url), "admin", "admin");
		Assertions.assertTrue(communicator.validateConnection());
	}

	protected void testDockerReceiveEmpty(String url) throws JMSException, IOException {
		this.logger.info("Testing MQ empty queue receiving: {}", url);
		MqCommunicator communicator = this.commsService.getCommunicator(URI.create(url), "admin", "admin");

		long startTime = System.currentTimeMillis();
		
		GenericDestination destination = new GenericDestination();
		destination.setQueueName("com.inteligr8.mq.test");
		try {
			communicator.receive(destination, 3000L);
			Assertions.fail();
		} catch (TimeoutException te) {
			long elapsedTime = System.currentTimeMillis() - startTime;
			if (elapsedTime < 3000L) {
				Assertions.fail("Timed out earlier than it should have: " + elapsedTime + " ms");
			} else if (elapsedTime > 3100L) {
				Assertions.fail("Timed out later than it should have: " + elapsedTime + " ms");
			}
		}
	}

	protected void testDockerSendReceive(String url) throws JMSException, IOException, TimeoutException, InterruptedException {
		this.logger.info("Testing MQ queue send/receive: {}", url);
		MqCommunicator communicator = this.commsService.getCommunicator(URI.create(url), "admin", "admin");
		
		GenericDestination destination = new GenericDestination();
		destination.setQueueName("com.inteligr8.mq.test");
		
		PreparedMessage<String> message = communicator.createPreparedMessage();
		message.setContent("{}", "application/json");
		message.setReplyToQueueName("com.inteligr8.mq.test.reply");

		OffsetDateTime beforeSend = OffsetDateTime.now();
		this.logger.debug("Timestamp before send: {}", beforeSend);
		
		DeliveredMessage<String> sentMessage = communicator.send(destination, message);
		Assertions.assertNotNull(sentMessage);
		Assertions.assertNotNull(sentMessage.getMessageId());
		Assertions.assertNotNull(sentMessage.getCorrelationId());
		
		DeliveredMessage<String> receivedMessage = communicator.receive(destination, 25L, sentMessage.getCorrelationId());
		Assertions.assertNotNull(receivedMessage);
		this.logger.debug("Timestamp of delivery: {}", receivedMessage.getDeliveryTime());
		
		Assertions.assertNotNull(receivedMessage.getMessageId());
		Assertions.assertEquals(sentMessage.getMessageId(), receivedMessage.getMessageId());
		Assertions.assertEquals(sentMessage.getCorrelationId(), receivedMessage.getCorrelationId());
		Assertions.assertEquals(message.getReplyToQueueName(), receivedMessage.getReplyToQueueName());
		Assertions.assertEquals("{}", receivedMessage.getContent());
		Assertions.assertNotNull(receivedMessage.getDeliveryTime());
		Assertions.assertFalse(receivedMessage.getDeliveryTime().isBefore(beforeSend));
		Assertions.assertFalse(receivedMessage.getDeliveryTime().isAfter(OffsetDateTime.now()));
		Assertions.assertNotNull(receivedMessage.getPriority());
	}

}
